from __future__ import annotations

import json
from http import HTTPStatus
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from baize.asgi import HTTPException, PlainTextResponse
from pydantic import ValidationError
from pydantic.json import pydantic_encoder
from typing_extensions import Literal

from .requests import request
from .responses import HttpResponse


class RequestValidationError(Exception):
    def __init__(
        self,
        validation_error: ValidationError,
        in_: Literal["path", "query", "header", "cookie", "body"],
    ) -> None:
        self.validation_error = validation_error
        self.in_ = in_

    def errors(self) -> List[Dict[str, Any]]:
        errors = self.validation_error.errors()
        for error in errors:
            error["in"] = self.in_
        return errors

    def json(self, *, indent: Union[None, int, str] = 2) -> str:
        return json.dumps(self.errors(), indent=indent, default=pydantic_encoder)

    @staticmethod
    def schema() -> Dict[str, Any]:
        return {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "loc": {
                        "title": "Location",
                        "description": "error field",
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "type": {
                        "title": "Type",
                        "description": "error type",
                        "type": "string",
                    },
                    "msg": {
                        "title": "Message",
                        "description": "error message",
                        "type": "string",
                    },
                    "ctx": {
                        "title": "Context",
                        "description": "error context",
                        "type": "string",
                    },
                    "in": {
                        "title": "In",
                        "type": "string",
                        "enum": ["path", "query", "header", "cookie", "body"],
                    },
                },
                "required": ["loc", "type", "msg"],
            },
        }


Error = TypeVar("Error", bound=BaseException)
ErrorView = Callable[[Error], Awaitable[HttpResponse]]


class ExceptionContextManager:
    def __init__(
        self,
        handlers: Mapping[Union[int, Type[BaseException]], ErrorView] = {},
    ) -> None:
        self._status_handlers: Dict[int, ErrorView] = {}
        self._exception_handlers: Dict[Type[BaseException], ErrorView] = {
            HTTPException: self.http_exception,
            RequestValidationError: self.validation_error,
        }
        for key, value in handlers.items():
            self.add_exception_handler(key, value)

    def add_exception_handler(
        self,
        exc_class_or_status_code: Union[int, Type[BaseException]],
        handler: Callable,
    ) -> None:
        if isinstance(exc_class_or_status_code, int):
            self._status_handlers[exc_class_or_status_code] = handler
        else:
            self._exception_handlers[exc_class_or_status_code] = handler

    def _lookup_exception_handler(self, exc: BaseException) -> Optional[Callable]:
        for cls in type(exc).__mro__:
            if cls in self._exception_handlers:
                return self._exception_handlers[cls]
        return None

    async def __aenter__(self) -> ExceptionContextManager:
        return self

    @overload
    async def __aexit__(self, exc_type: None, exc: None, tb: None) -> None:
        ...

    @overload
    async def __aexit__(
        self, exc_type: Type[BaseException], exc: BaseException, tb: TracebackType
    ) -> bool:
        ...

    async def __aexit__(self, exc_type, exc, tb):
        if exc_type is None:
            return None
        else:
            handler = None
            if isinstance(exc, HTTPException):
                handler = self._status_handlers.get(exc.status_code)
            if handler is None:
                handler = self._lookup_exception_handler(exc)
            if handler is None:
                return False
            response = await handler(exc)
            await response(request._scope, request._receive, request._send)
            return True

    @staticmethod
    async def http_exception(exc: HTTPException) -> HttpResponse:
        if exc.status_code in {204, 304}:
            return HttpResponse(status_code=exc.status_code, headers=exc.headers)
        else:
            return request.app.response_converter(
                exc.content
                if exc.content is not None
                else HTTPStatus(exc.status_code).description,
                exc.status_code,
                exc.headers,
            )

    async def validation_error(self, exc: RequestValidationError) -> HttpResponse:
        if exc.in_ == "path":
            return await self.http_exception(HTTPException(status_code=404))
        else:
            return PlainTextResponse(
                exc.json(), status_code=422, media_type="application/json"
            )



from __future__ import annotations

import copy
import dataclasses
import functools
import inspect
import traceback
from functools import reduce
from pathlib import PurePath
from types import AsyncGeneratorType
from typing import (
    Any,
    Awaitable,
    Callable,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
)

from baize.asgi import Receive, Scope, Send
from baize.datastructures import URL
from baize.typing import ASGIApp
from typing_extensions import Literal

from .debug import DebugMiddleware
from .exceptions import ErrorView, ExceptionContextManager, HTTPException
from .requests import HttpRequest, WebSocket, request_var, websocket_var
from .responses import (
    FileResponse,
    HttpResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    SendEventResponse,
    convert_response,
)
from .routing import BaseRoute, NoMatchFound, Router
from .templates import BaseTemplates
from .utils import F, ImmutableAttribute, State

NoArgumentCallable = Union[Callable[[], Any], Callable[[], Awaitable[Any]]]

T_NoArgumentCallable = TypeVar("T_NoArgumentCallable", bound=NoArgumentCallable)


@dataclasses.dataclass
class Lifespan:
    on_startup: List[NoArgumentCallable] = dataclasses.field(default_factory=list)
    on_shutdown: List[NoArgumentCallable] = dataclasses.field(default_factory=list)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handle ASGI lifespan messages, which allows us to manage application
        startup and shutdown events.
        """
        message = await receive()
        assert message["type"] == "lifespan.startup"
        try:
            for handler in self.on_startup:
                result = handler()
                if inspect.isawaitable(result):
                    await result
        except BaseException:
            msg = traceback.format_exc()
            await send({"type": "lifespan.startup.failed", "message": msg})
            raise
        await send({"type": "lifespan.startup.complete"})

        message = await receive()
        assert message["type"] == "lifespan.shutdown"
        try:
            for handler in self.on_shutdown:
                result = handler()
                if inspect.isawaitable(result):
                    await result
        except BaseException:
            msg = traceback.format_exc()
            await send({"type": "lifespan.shutdown.failed", "message": msg})
            raise
        await send({"type": "lifespan.shutdown.complete"})


@dataclasses.dataclass
class FactoryClass:
    http: Type[HttpRequest] = HttpRequest
    websocket: Type[WebSocket] = WebSocket


class Index:
    debug: ImmutableAttribute[bool] = ImmutableAttribute()
    state: ImmutableAttribute[State] = ImmutableAttribute()

    def __init__(
        self,
        *,
        debug: bool = False,
        templates: Optional[BaseTemplates] = None,
        on_startup: List[NoArgumentCallable] = [],
        on_shutdown: List[NoArgumentCallable] = [],
        routes: Iterable[BaseRoute] = [],
        exception_handlers: Mapping[int | Type[BaseException], ErrorView] = {},
        factory_class: FactoryClass = FactoryClass(),
        response_converters: Mapping[type, Callable[..., HttpResponse]] = {},
    ) -> None:
        self.debug = debug
        self.state = State()
        self.response_converter = create_response_converter(response_converters)
        self.should_exit = False
        self.factory_class = factory_class
        self.templates = templates
        self.router = Router(routes)
        self.lifespan = Lifespan(copy.copy(on_startup), copy.copy(on_shutdown))
        self.exception_contextmanager = ExceptionContextManager(exception_handlers)
        self._asgi_middlewares: List[F] = []
        self.app_with_middlewares = self.build_app_with_middlewares()

    def build_app_with_middlewares(self) -> ASGIApp:
        internal_middlewares: List[F] = []
        if self.debug:
            internal_middlewares.append(F(DebugMiddleware))

        return reduce(
            lambda a, m: m(a),
            reversed([*self._asgi_middlewares, *internal_middlewares]),
            self.app,
        )

    def add_middleware(self, middleware_class: type, **options: Any) -> None:
        """
        Add ASGI middleware
        """
        self._asgi_middlewares.append(F(middleware_class, **options))
        self.app_with_middlewares = self.build_app_with_middlewares()

    def add_exception_handler(
        self, exc_class_or_status_code: int | Type[Exception], handler: ErrorView
    ) -> None:
        self.exception_contextmanager.add_exception_handler(
            exc_class_or_status_code, handler
        )

    def exception_handler(
        self, exc_class_or_status_code: int | Type[Exception]
    ) -> Callable[[ErrorView], ErrorView]:
        def decorator(func: ErrorView) -> ErrorView:
            self.add_exception_handler(exc_class_or_status_code, func)
            return func

        return decorator

    def on_startup(self, func: T_NoArgumentCallable) -> T_NoArgumentCallable:
        self.lifespan.on_startup.append(func)
        return func

    def on_shutdown(self, func: T_NoArgumentCallable) -> T_NoArgumentCallable:
        self.lifespan.on_shutdown.append(func)
        return func

    async def app(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope_type: Literal["lifespan", "http", "websocket"] = scope["type"]
        return await getattr(self, scope_type)(scope, receive, send)

    async def http(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = self.factory_class.http(scope, receive, send)
        token = request_var.set(request)
        try:
            async with self.exception_contextmanager:
                try:
                    path_params, handler = self.router.search("http", request["path"])
                    request["path_params"] = path_params
                    response = convert_response(await handler())
                except NoMatchFound:
                    raise HTTPException(404)
                else:
                    return await response(scope, receive, send)
        finally:
            await request.close()
            request_var.reset(token)

    async def websocket(self, scope: Scope, receive: Receive, send: Send) -> None:
        websocket = self.factory_class.websocket(scope, receive, send)
        token = websocket_var.set(websocket)
        try:
            path_params, handler = self.router.search("websocket", websocket["path"])
            websocket["path_params"] = path_params
        except NoMatchFound:
            return await websocket.close(1001)
        else:
            return await handler()
        finally:
            websocket_var.reset(token)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self

        await self.app_with_middlewares(scope, receive, send)


def create_response_converter(
    converters: Mapping[type, Callable[..., HttpResponse]]
) -> functools._SingleDispatchCallable[HttpResponse]:
    """
    Create a converter for convert response.
    """

    @functools.singledispatch
    def response_converter(*args: Any) -> HttpResponse:
        raise TypeError(
            f"Cannot find response_converter handler for this type: {type(args[0])}"
        )

    def _none(ret: Type[None]) -> NoReturn:
        raise TypeError(
            "Get 'None'. Maybe you need to add a return statement to the function."
        )

    response_converter.register(type(None), _none)
    response_converter.register(HttpResponse, lambda x: x)
    response_converter.register(dict, JSONResponse)
    response_converter.register(list, JSONResponse)
    response_converter.register(tuple, JSONResponse)
    response_converter.register(bytes, PlainTextResponse)
    response_converter.register(str, PlainTextResponse)
    response_converter.register(AsyncGeneratorType, SendEventResponse)
    response_converter.register(
        PurePath,
        lambda filepath, download_name=None: FileResponse(
            str(filepath), download_name=download_name
        ),
    )
    response_converter.register(URL, RedirectResponse)

    for type_, converter in converters.items():
        response_converter.register(type_, converter)

    return response_converter



from typing import Any, Dict, Generic, NoReturn, Tuple, TypeVar, overload

T = TypeVar("T")


class Singleton(type):
    def __init__(
        cls,
        name: str,
        bases: Tuple[type],
        namespace: Dict[str, Any],
    ) -> None:
        cls.instance = None
        super().__init__(name, bases, namespace)

    def __call__(cls, *args, **kwargs) -> Any:
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)
        return cls.instance


class ImmutableAttribute(Generic[T]):
    def __set_name__(self, owner: object, name: str) -> None:
        self.public_name = name
        self.private_name = "_" + name

    @overload
    def __get__(self, instance: None, cls: type = None) -> NoReturn:
        ...

    @overload
    def __get__(self, instance: object, cls: type = None) -> T:
        ...

    def __get__(self, instance, cls):
        if instance is None:
            raise AttributeError(
                f"{instance.__class__.__name__} has no attribute '{self.public_name}'"
            ) from None
        else:
            return getattr(instance, self.private_name)

    def __set__(self, instance: object, value: T) -> None:
        if hasattr(instance, self.private_name):
            raise RuntimeError(
                f"{instance.__class__.__name__}.{self.public_name} is immutable"
            ) from None
        setattr(instance, self.private_name, value)

    def __delete__(self, instance: object) -> None:
        raise RuntimeError(
            f"{instance.__class__.__name__}.{self.public_name} is immutable"
        ) from None



from __future__ import annotations

import typing

T = typing.TypeVar("T", bound=typing.Callable)


def merge_openapi_info(
    operation_info: typing.Dict[str, typing.Any],
    more_info: typing.Dict[str, typing.Any],
) -> typing.Dict[str, typing.Any]:
    for key, value in more_info.items():
        if key in operation_info:
            if isinstance(operation_info[key], typing.Sequence):
                operation_info[key] = _ = list(operation_info[key])
                _.extend(value)
                continue
            elif isinstance(operation_info[key], dict):
                operation_info[key] = merge_openapi_info(operation_info[key], value)
                continue
        operation_info[key] = value
    return operation_info


def describe_extra_docs(handler: T, info: typing.Dict[str, typing.Any]) -> T:
    """
    describe more openapi info in HTTP handler

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.3.md#operationObject
    """
    if isinstance(handler, type):
        for method in getattr(handler, "__methods__"):
            handler_method = getattr(handler, method.lower())
            __extra_docs__ = merge_openapi_info(
                getattr(handler_method, "__docs_extra__", {}), info
            )
            setattr(handler_method, "__docs_extra__", __extra_docs__)
    else:
        __extra_docs__ = merge_openapi_info(
            getattr(handler, "__docs_extra__", {}), info
        )
        setattr(handler, "__docs_extra__", __extra_docs__)
    return handler



from __future__ import annotations

from .applications import Index
from .exceptions import HTTPException
from .parameters.field_functions import (
    Body,
    Cookie,
    Depends,
    Header,
    Path,
    Query,
    Request,
)
from .requests import request, websocket
from .responses import (
    FileResponse,
    HTMLResponse,
    HttpResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    SendEventResponse,
    ServerSentEvent,
    StreamResponse,
    TemplateResponse,
)
from .routing import HttpRoute, Routes, SocketRoute
from .views import HttpView, SocketView, required_method

__all__ = [
    "Index",
    "request",
    "websocket",
    "HTTPException",
    "Body",
    "Cookie",
    "Header",
    "Path",
    "Query",
    "Request",
    "Depends",
    "HttpResponse",
    "FileResponse",
    "HTMLResponse",
    "JSONResponse",
    "PlainTextResponse",
    "RedirectResponse",
    "ServerSentEvent",
    "SendEventResponse",
    "StreamResponse",
    "TemplateResponse",
    "required_method",
    "HttpView",
    "SocketView",
    "Routes",
    "HttpRoute",
    "SocketRoute",
]



import functools
import re
from typing import Any, Awaitable, Callable, Dict, Iterable, Pattern, TypeVar

from .. import request
from ..responses import convert_response

T = TypeVar("T")

View = Callable[..., Awaitable[T]]


def allow_cors(
    allow_origins: Iterable[Pattern] = (re.compile(".*"),),
    allow_methods: Iterable[str] = (
        "GET",
        "POST",
        "PUT",
        "PATCH",
        "DELETE",
        "HEAD",
        "OPTIONS",
        "TRACE",
    ),
    allow_headers: Iterable[str] = (),
    expose_headers: Iterable[str] = (),
    allow_credentials: bool = False,
    max_age: int = 600,
) -> Callable[[View], View]:
    """
    Cross-Origin Resource Sharing
    """

    config_dict: Dict[str, str] = {
        "Access-Control-Allow-Methods": ", ".join(allow_methods),
        "Access-Control-Allow-Headers": ", ".join(
            {"Accept", "Accept-Language", "Content-Language", "Content-Type"}
            | set(allow_headers)
        ),
        "Access-Control-Expose-Headers": ", ".join(expose_headers),
        "Access-Control-Allow-Credentials": "true" if allow_credentials else "false",
        "Access-Control-Max-Age": str(max_age),
    }
    config_dict = {k: v for k, v in config_dict.items() if v}

    def decorator(endpoint: View) -> View:
        @functools.wraps(endpoint)
        async def cors_wrapper(*args: Any, **kwargs: Any) -> Any:
            origin = request.headers.get("origin", None)
            if origin and any(
                origin_pattern.fullmatch(origin) for origin_pattern in allow_origins
            ):
                response = convert_response(await endpoint(*args, **kwargs))
                response.headers.update(config_dict)
                response.headers["Access-Control-Allow-Origin"] = origin
                return response
            else:
                return await endpoint(*args, **kwargs)

        return cors_wrapper

    return decorator


CORSMiddleware = allow_cors



from __future__ import annotations

import asyncio
import copy
import inspect
import json
import operator
import typing
from functools import reduce
from hashlib import md5
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, TypeVar

from typing_extensions import Literal, TypedDict

if typing.TYPE_CHECKING:
    from ..applications import Index
    from ..requests import HttpRequest

from ..exceptions import RequestValidationError
from ..requests import request
from ..responses import HTMLResponse, JSONResponse
from ..routing import HttpRoute, Routes
from . import specification as spec
from .extra_docs import merge_openapi_info
from .schema import schema_parameter, schema_request_body, schema_response


class TagDetailOptional(TypedDict, total=False):
    paths: Sequence[str]


class TagDetail(TagDetailOptional):
    description: str


class OpenAPI:
    def __init__(
        self,
        info: spec.Info = {"title": "IndexPy API", "version": "1.0.0"},
        security_schemes: Dict[str, spec.SecurityScheme | spec.Reference] = {},
        *,
        tags: Dict[str, TagDetail] = {},
        template_name: Literal["redoc", "swagger", "rapidoc"] = "swagger",
        template: str = "",
    ) -> None:
        if template == "":
            template = (
                Path(__file__).absolute().parent / "templates" / f"{template_name}.html"
            ).read_text(encoding="utf8")

        self.html_template = template

        self.openapi = spec.OpenAPI(
            openapi="3.0.3",
            info=info,
            paths={},
            tags=[
                {"name": tag_name, "description": tag_info.get("description", "")}
                for tag_name, tag_info in tags.items()
            ],
            components={
                "schemas": {
                    "RequestValidationError": typing.cast(
                        spec.Schema, RequestValidationError.schema()
                    )
                },
            },
        )
        if security_schemes:
            self.openapi.get("components", {})["securitySchemes"] = security_schemes
        self.path2tag: Dict[str, List[str]] = {}
        for tag_name, tag_info in tags.items():
            for path in tag_info.get("paths", []):
                self.path2tag.setdefault(path, []).append(tag_name)
        self.definitions: dict = {}

    def _generate_paths(self, app: Index) -> Tuple[spec.Paths, dict]:
        _definitions: dict = {}
        update_definitions = lambda path_item, x: _definitions.update(x) or path_item
        return {
            path: openapi_path_item
            for path, openapi_path_item in (
                (
                    path_format,
                    update_definitions(*self._generate_path(handler, path_format)),
                )
                for path_format, handler in app.router.http_tree.iterator()
            )
            if openapi_path_item
        }, _definitions

    def _generate_path(self, view: Any, path: str) -> Tuple[spec.PathItem, dict]:
        """
        Generate documents under a path
        """
        _definitions: dict = {}
        update_definitions = lambda path_item, x: _definitions.update(x) or path_item
        if hasattr(view, "__methods__"):
            result = clear_empty(
                {
                    method: update_definitions(
                        *self._generate_method(getattr(view, method), path)
                    )
                    for method in (
                        method.lower()
                        for method in typing.cast(Iterable[str], view.__methods__)
                        if method.lower() != "options"
                    )
                }
            )
        elif hasattr(view, "__method__"):
            result = clear_empty(
                {
                    typing.cast(str, view.__method__).lower(): update_definitions(
                        *self._generate_method(view, path)
                    ),
                }
            )
        else:
            result = {}

        return typing.cast(spec.PathItem, result), _definitions

    def _generate_method(self, func: Any, path: str) -> Tuple[spec.Operation, dict]:
        result: Dict[str, Any] = {}
        _definitions: dict = {}
        update_definitions = lambda path_item, x: _definitions.update(x) or path_item

        # This is mypy check error, if you use pyright/pylance, this is all fine.
        if hasattr(func, "__summary__") and isinstance(func.__summary__, str):  # type: ignore
            result["summary"] = func.__summary__  # type: ignore
        if hasattr(func, "__description__") and isinstance(func.__description__, str):  # type: ignore
            result["description"] = func.__description__  # type: ignore

        if isinstance(func.__doc__, str):
            clean_doc = inspect.cleandoc(func.__doc__)
            if "summary" not in result and "description" not in result:
                result.update(
                    zip(("summary", "description"), clean_doc.split("\n\n", 1))
                )
            elif "description" not in result:
                result["description"] = clean_doc
            else:
                result["summary"] = ""

        # generate params schema
        parameters = reduce(
            operator.add,
            map(
                lambda key: schema_parameter(
                    create_model(getattr(func, "__docs_parameters__", {}).get(key, [])),
                    typing.cast(Literal["path", "query", "header", "cookie"], key),
                ),
                ["path", "query", "header", "cookie"],
            ),
        )
        result["parameters"] = parameters

        # generate request body schema
        request_body = update_definitions(
            *schema_request_body(
                create_model(getattr(func, "__docs_request_body__", []))
            )
        )
        result["requestBody"] = request_body

        # generate responses schema
        responses: spec.Responses = {}
        if parameters or request_body:
            responses["422"] = {
                "content": {
                    "application/json": {
                        "schema": {
                            "$ref": "#/components/schemas/RequestValidationError"
                        }
                    }
                },
                "description": "Failed to verify request parameters",
            }
        for status, info in getattr(func, "__docs_responses__", {}).items():
            _ = responses[status] = copy.copy(info)
            if _.get("content") is not None:
                _["content"] = update_definitions(*schema_response(_["content"]))

        result["responses"] = responses

        # set path tags
        if result and path in self.path2tag:
            result["tags"] = self.path2tag[path]

        result["tags"] = getattr(func, "__tags__", []) + result.get("tags", [])

        # merge user custom operation info
        return (
            typing.cast(
                spec.Operation,
                merge_openapi_info(
                    clear_empty(result), getattr(func, "__docs_extra__", {})
                ),
            ),
            _definitions,
        )

    def create_docs(self, request: HttpRequest) -> spec.OpenAPI:
        openapi = copy.deepcopy(self.openapi)
        openapi["servers"] = [
            {
                "url": f"{request.url.scheme}://{request.url.netloc}",
                "description": "Current server",
            },
            spec.Server(
                url="{scheme}://{address}/",
                description="Custom API Server Host",
                variables={
                    "scheme": {
                        "default": request.url.scheme,
                        "enum": ["http", "https"],
                        "description": "http or https",
                    },
                    "address": {
                        "default": request.url.netloc,
                        "description": "api server's host[:port]",
                    },
                },
            ),
        ]
        openapi["paths"], definitions = copy.deepcopy(self._generate_paths(request.app))
        for path_item in openapi["paths"].values():
            for operation in filter(lambda x: isinstance(x, dict), path_item.values()):
                operation = typing.cast(spec.Operation, operation)
                if "responses" not in operation:
                    operation["responses"] = {}
        openapi.get("components", {}).setdefault("schemas", {}).update(definitions)
        return openapi

    @property
    def routes(self) -> Routes:
        async def redirect():
            return request.url.replace(path=request.url.path + "/")

        async def template():
            return HTMLResponse(self.html_template)

        async def json_docs():
            openapi = self.create_docs(request)
            return JSONResponse(
                openapi, headers={"hash": md5(json.dumps(openapi).encode()).hexdigest()}
            )

        async def heartbeat():
            async def g():
                openapi = self.create_docs(request)
                yield {
                    "id": md5(json.dumps(openapi).encode()).hexdigest(),
                    "data": json.dumps(openapi),
                }
                while not request.app.should_exit:
                    await asyncio.sleep(1)

            return g()

        return Routes(
            HttpRoute("", redirect),
            HttpRoute("/", template),
            HttpRoute("/json", json_docs),
            HttpRoute("/heartbeat", heartbeat),
        )


T_Dict = TypeVar("T_Dict", bound=Dict)


def clear_empty(d: T_Dict) -> T_Dict:
    return typing.cast(T_Dict, {k: v for k, v in d.items() if v})


def create_model(bases: List[type]) -> Optional[type]:
    if bases:
        return type("T_SchemaModel", tuple(bases), {})
    else:
        return None



from __future__ import annotations

import asyncio
import contextvars
import functools
import typing
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Generator,
    Iterator,
    TypeVar,
    overload,
)

from .utils import is_async_gen_callable, is_coroutine_callable, is_gen_callable

T = TypeVar("T")


async def run_in_threadpool(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """
    Run function in threadpool.
    """
    loop = asyncio.get_running_loop()
    child = functools.partial(func, *args, **kwargs)
    context = contextvars.copy_context()
    return typing.cast(T, await loop.run_in_executor(None, context.run, child))


class _StopIteration(Exception):
    pass


def _next(iterator: Iterator[T]) -> T:
    try:
        return next(iterator)
    except StopIteration:
        raise _StopIteration


async def iterate_in_threadpool(iterator: Iterator[T]) -> AsyncGenerator[T, None]:
    """
    Convert iterator to async generator.
    """
    loop = asyncio.get_running_loop()
    context = contextvars.copy_context()
    while True:
        try:
            child = functools.partial(_next, iterator)
            yield typing.cast(T, await loop.run_in_executor(None, context.run, child))
        except _StopIteration:
            break


@overload
def always_async(  # type: ignore
    func: Callable[..., Generator[T, None, None]]
) -> Callable[..., AsyncGenerator[T, None]]:
    ...


@overload
def always_async(func: Callable[..., T]) -> Callable[..., Awaitable[T]]:
    ...


def always_async(func):
    if is_coroutine_callable(func) or is_async_gen_callable(func):
        return func

    if is_gen_callable(func):

        @functools.wraps(func)
        async def wrapper_gen(*args, **kwargs):
            generator = func(*args, **kwargs)
            async for item in iterate_in_threadpool(generator):
                yield item

        return wrapper_gen

    else:

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            return await run_in_threadpool(func, *args, **kwargs)

        return wrapper



import operator
from typing import Callable, Dict, Iterator, MutableMapping, TypeVar

KT = TypeVar("KT")
VT = TypeVar("VT")


class RegisterDict(MutableMapping[KT, VT]):
    """
    A mapping that can be used to register items.
    """

    def __init__(self) -> None:
        self._store: Dict[KT, VT] = {}

    def __getitem__(self, key: KT) -> VT:
        return self._store[key]

    def __setitem__(self, key: KT, value: VT) -> None:
        self._store[key] = value

    def __delitem__(self, key: KT) -> None:
        del self._store[key]

    def __iter__(self) -> Iterator[KT]:
        return iter(self._store)

    def __len__(self) -> int:
        return len(self._store)

    def register(self, key: KT) -> Callable[[VT], VT]:
        return lambda value: operator.setitem(self, key, value) or value



def bind_contextvar(contextvar):
    class ContextVarBind:
        __slots__ = ()

        def __getattr__(self, name):
            return getattr(contextvar.get(), name)

        def __setattr__(self, name, value):
            setattr(contextvar.get(), name, value)

        def __delattr__(self, name):
            delattr(contextvar.get(), name)

        def __getitem__(self, index):
            return contextvar.get()[index]

        def __setitem__(self, index, value):
            contextvar.get()[index] = value

        def __delitem__(self, index):
            del contextvar.get()[index]

    return ContextVarBind()



from __future__ import annotations

from .contextvars import bind_contextvar
from .importer import import_from_string, import_module
from .inspect import (
    get_object_filepath,
    get_raw_handler,
    is_async_gen_callable,
    is_coroutine_callable,
    is_gen_callable,
    safe_issubclass,
)
from .objects import ImmutableAttribute, Singleton
from .pipe import FF, F
from .state import State

__all__ = [
    "Singleton",
    "ImmutableAttribute",
    "bind_contextvar",
    "FF",
    "F",
    "State",
    "is_async_gen_callable",
    "is_coroutine_callable",
    "is_gen_callable",
    "safe_issubclass",
    "get_raw_handler",
    "get_object_filepath",
    "Singleton",
    "import_module",
    "import_from_string",
    "ImmutableAttribute",
]



from __future__ import annotations

import typing
from http import HTTPStatus

from baize import asgi as baize_asgi
from pydantic import BaseModel, create_model
from pydantic.json import pydantic_encoder
from pydantic.typing import display_as_type

from .requests import request
from .utils import safe_issubclass

__all__ = [
    "convert_response",
    "HttpResponse",
    "FileResponse",
    "HTMLResponse",
    "JSONResponse",
    "PlainTextResponse",
    "RedirectResponse",
    "SendEventResponse",
    "ServerSentEvent",
    "StreamResponse",
    "TemplateResponse",
]

HttpResponse = baize_asgi.Response
ServerSentEvent = baize_asgi.ServerSentEvent


class JSONResponse(baize_asgi.JSONResponse):
    def __init__(
        self,
        content: typing.Any,
        status_code: int = 200,
        headers: typing.Mapping[str, str] = None,
        **kwargs: typing.Any,
    ) -> None:
        json_kwargs: typing.Dict[str, typing.Any] = {
            "default": pydantic_encoder,
        }
        json_kwargs.update(**kwargs)
        super().__init__(
            content, status_code=status_code, headers=headers, **json_kwargs
        )

    def __class_getitem__(cls, parameters: typing.Tuple[typing.Any, ...]):
        """
        Use JSONResponse[status, headers, content] to describe response
        """
        if isinstance(parameters, tuple):
            assert len(parameters) in (2, 3)
            if len(parameters) == 2:
                (status_code, headers), content = parameters, {}
            else:
                status_code, headers, content = parameters
        else:
            status_code, headers, content = parameters, {}, {}
        assert isinstance(status_code, int) or status_code == "default"

        docs: typing.Dict[str, typing.Any] = {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
            }
        }
        if headers:
            docs[str(status_code)]["headers"] = headers

        if content:
            if isinstance(content, dict) or (
                getattr(content, "__origin__", None) is None
                and safe_issubclass(content, BaseModel)
            ):
                real_content = content
            else:
                real_content = create_model(
                    f"ParsingModel[{display_as_type(content)}]", __root__=(content, ...)
                )
            docs[str(status_code)]["content"] = real_content

        return docs


class FileResponse(baize_asgi.FileResponse):
    def __class_getitem__(cls, parameters: typing.Tuple[typing.Any, ...]):
        """
        Use FileResponse[content_type, headers] to describe response
        """
        if isinstance(parameters, tuple):
            content_type, headers = parameters
        else:
            content_type, headers = parameters, {}
        assert isinstance(content_type, str)
        docs = {
            "200": {
                "description": HTTPStatus.OK.description,
                "content": {
                    content_type: {"schema": {"type": "string", "format": "binary"}}
                },
            },
            "206": {
                "description": HTTPStatus.PARTIAL_CONTENT.description,
                "content": {
                    content_type: {"schema": {"type": "string", "format": "binary"}}
                },
            },
        }
        if headers:
            docs["200"]["headers"] = headers
            docs["206"]["headers"] = headers
        return docs


class PlainTextResponse(baize_asgi.PlainTextResponse):
    def __class_getitem__(cls, parameters: typing.Tuple[typing.Any, ...]):
        """
        Use PlainTextResponse[status, headers] to describe response
        """
        if isinstance(parameters, tuple):
            status_code, headers = parameters
        else:
            status_code, headers = parameters, {}
        assert isinstance(status_code, int)
        docs = {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "content": {"text/plain": {"schema": {"type": "string"}}},
            }
        }
        if headers:
            docs[str(status_code)]["headers"] = headers
        return docs


class HTMLResponse(baize_asgi.HTMLResponse):
    def __class_getitem__(cls, parameters: typing.Tuple[typing.Any, ...]):
        """
        Use HTMLResponse[status, headers] to describe response
        """
        if isinstance(parameters, tuple):
            status_code, headers = parameters
        else:
            status_code, headers = parameters, {}
        assert isinstance(status_code, int)
        docs = {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "content": {"text/html": {"schema": {"type": "string"}}},
            }
        }
        if headers:
            docs[str(status_code)]["headers"] = headers
        return docs


class RedirectResponse(baize_asgi.RedirectResponse):
    def __class_getitem__(cls, parameters: typing.Tuple[typing.Any, ...]):
        """
        Use RedirectResponse[status, headers] to describe response
        """
        if isinstance(parameters, tuple):
            status_code, headers = parameters
        else:
            status_code, headers = parameters, {}
        assert isinstance(status_code, int)
        docs = {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "headers": {
                    "Location": {"schema": {"type": "string"}},
                    **headers,
                },
            }
        }
        return docs


class SendEventResponse(baize_asgi.SendEventResponse):
    def __class_getitem__(cls, parameters: typing.Tuple[typing.Any, ...]):
        """
        Use SendEventResponse[status, headers] to describe response
        """
        if isinstance(parameters, tuple):
            status_code, headers = parameters
        else:
            status_code, headers = parameters, {}
        assert isinstance(status_code, int)

        # TODO: Follow the spec
        # https://github.com/OAI/OpenAPI-Specification/issues/396
        docs = {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "content": {"text/event-stream": {"schema": {"type": "string"}}},
            }
        }
        if headers:
            docs[str(status_code)]["headers"] = headers
        return docs


class StreamResponse(baize_asgi.StreamResponse):
    def __class_getitem__(cls, parameters: typing.Tuple[typing.Any, ...]):
        """
        Use StreamResponse[status, headers] to describe response
        """
        if isinstance(parameters, tuple):
            status_code, headers = parameters
        else:
            status_code, headers = parameters, {}
        assert isinstance(status_code, int)

        # TODO: Follow the spec
        # https://github.com/OAI/OpenAPI-Specification/issues/1576
        docs = {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "headers": {
                    "Transfer-Encoding": {
                        "schema": {"type": "string"},
                        "description": "chunked",
                    },
                    **headers,
                },
            }
        }
        return docs


def TemplateResponse(
    name: str,
    context: typing.Mapping[str, typing.Any],
    status_code: int = 200,
    headers: typing.Mapping[str, str] = None,
    media_type: str = None,
    charset: str = None,
) -> HttpResponse:
    templates = request.app.templates
    if templates is None:
        raise RuntimeError(
            "You must assign a value to `app.templates` to use TemplateResponse"
        )

    return templates.TemplateResponse(
        name, context, status_code, headers, media_type, charset
    )


def convert_response(response: typing.Any) -> HttpResponse:
    """
    shortcut for automatic

    Example:

        response = convert(response)

    It is equivalent to:

        if isinstance(response, tuple):
            response = automatic(*response)
        else:
            response = automatic(response)

    """
    if isinstance(response, tuple):
        return request.app.response_converter(*response)
    else:
        return request.app.response_converter(response)



from __future__ import annotations

import inspect
from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Type, Union
from typing import cast as typing_cast

from baize.datastructures import ContentType
from pydantic import BaseModel
from typing_extensions import Literal, get_args, get_type_hints

from .. import request
from . import specification as spec
from .types import UploadFile

REF_TEMPLATE = "#/components/schemas/{model}"


def schema_parameter(
    m: Optional[Type[BaseModel]],
    position: Literal["path", "query", "header", "cookie"],
) -> List[spec.Parameter]:

    if m is None:
        return []

    _schemas = deepcopy(m.schema(ref_template=REF_TEMPLATE))
    properties: Dict[str, Any] = _schemas["properties"]
    required = _schemas.get("required", ())

    return [
        {
            "in": position,
            "name": name,
            "description": schema.pop("description", ""),
            "required": name in required,  # type: ignore
            "schema": schema,
            "deprecated": schema.pop("deprecated", False),
        }
        for name, schema in properties.items()
    ]


def schema_request_body(
    body: Type[BaseModel] = None,
) -> Tuple[Optional[spec.RequestBody], dict]:
    if body is None:
        return None, {}

    schema: Dict = deepcopy(body.schema(ref_template=REF_TEMPLATE))
    schema.pop("title")
    definitions = schema.pop("definitions", {})
    content_types = [
        str(v)
        for v in get_args(get_type_hints(request.data, include_extras=True)["return"])
        if isinstance(v, ContentType)
    ]

    for field in body.__fields__.values():
        if inspect.isclass(field.type_) and issubclass(field.type_, UploadFile):
            content_types = ["multipart/form-data"]

    return {
        "required": True,
        "content": {
            content_type: {"schema": deepcopy(schema)} for content_type in content_types
        },
    }, definitions


def schema_response(
    content: Union[Type[BaseModel], Dict[str, spec.MediaType]]
) -> Tuple[Dict[str, spec.MediaType], dict]:
    if isinstance(content, dict):
        return content, {}
    schema = deepcopy(content.schema(ref_template=REF_TEMPLATE))
    schema.pop("title")
    definitions = schema.pop("definitions", {})
    return {
        "application/json": {"schema": typing_cast(spec.Schema, schema)}
    }, definitions



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['indexpy',
 'indexpy.middlewares',
 'indexpy.openapi',
 'indexpy.parameters',
 'indexpy.routing',
 'indexpy.routing.extensions',
 'indexpy.utils']

package_data = \
{'': ['*'], 'indexpy.openapi': ['templates/*']}

install_requires = \
['baize>=0.12.0,<0.13.0',
 'pydantic>=1.8,<2.0',
 'typing-extensions>=3.10.0,<4.0.0']

extras_require = \
{'cli': ['click>=8.0,<9.0']}

entry_points = \
{'console_scripts': ['index-cli = indexpy.cli:index_cli']}

setup_kwargs = {
    'name': 'index.py',
    'version': '0.21.6',
    'description': 'An easy-to-use high-performance asynchronous web framework.',
    'long_description': '<div align="center">\n\n<img src="https://raw.githubusercontent.com/abersheeran/index.py/master/docs/img/index-py.png" />\n\n<p>\n中文\n|\n<a href="https://github.com/abersheeran/index.py/tree/master/README-en.md">English</a>\n</p>\n\n<p>\n<a href="https://github.com/abersheeran/index.py/actions?query=workflow%3ATest">\n<img src="https://github.com/abersheeran/index.py/workflows/Test/badge.svg" alt="Github Action Test" />\n</a>\n\n<a href="https://app.codecov.io/gh/abersheeran/index.py/">\n<img alt="Codecov" src="https://img.shields.io/codecov/c/github/abersheeran/index.py">\n</a>\n</p>\n\n<p>\n<a href="https://github.com/abersheeran/index.py/actions?query=workflow%3A%22Publish+PyPi%22">\n<img src="https://github.com/abersheeran/index.py/workflows/Publish%20PyPi/badge.svg" alt="Publish PyPi" />\n</a>\n\n<a href="https://pypi.org/project/index.py/">\n<img src="https://img.shields.io/pypi/v/index.py" alt="PyPI" />\n</a>\n\n<a href="https://pepy.tech/project/index-py">\n<img src="https://static.pepy.tech/personalized-badge/index-py?period=total&units=international_system&left_color=black&right_color=blue&left_text=PyPi%20Downloads" alt="Downloads">\n</a>\n</p>\n\n<p>\n<img src="https://img.shields.io/pypi/pyversions/index.py" alt="PyPI - Python Version" />\n</p>\n\n一个易用的高性能异步 web 框架。\n\n<a href="https://index-py.aber.sh/stable/">Index.py 文档</a>\n\n</div>\n\n---\n\nIndex.py 实现了 [ASGI3](http://asgi.readthedocs.io/en/latest/) 接口，并使用 Radix Tree 进行路由查找。是[最快的 Python web 框架之一](https://github.com/the-benchmarker/web-frameworks)。一切特性都服务于快速开发高性能的 Web 服务。\n\n- 大量正确的类型注释\n- 灵活且高效的路由系统\n- 可视化 API 接口与在线调试\n- 支持 [Server-sent events](https://developer.mozilla.org/zh-CN/docs/Web/API/Server-sent_events/Using_server-sent_events) 与 WebSocket\n- 自带一键部署命令 (基于 uvicorn 与 gunicorn)\n- 可使用任何可用的 ASGI 生态\n\n## Install\n\n```bash\npip install -U index.py\n```\n\n或者直接从 Github 上安装最新版本（不稳定）\n\n```bash\npip install -U git+https://github.com/abersheeran/index.py@setup.py\n```\n\n中国大陆内的用户可从 Gitee 上的镜像仓库拉取\n\n```bash\npip install -U git+https://gitee.com/abersheeran/index.py.git@setup.py\n```\n',
    'author': 'abersheeran',
    'author_email': 'me@abersheeran.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abersheeran/index.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import inspect
from pathlib import Path
from typing import Any, AsyncGenerator, Awaitable, Callable, Generator, TypeVar

from typing_extensions import TypeGuard

_ClassType = TypeVar("_ClassType", bound=type)


def safe_issubclass(__cls: Any, __class: _ClassType) -> TypeGuard[_ClassType]:
    return isinstance(__cls, type) and issubclass(__cls, __class)


def is_coroutine_callable(call: Callable) -> TypeGuard[Callable[..., Awaitable]]:
    if inspect.isroutine(call):
        return inspect.iscoroutinefunction(call)
    if inspect.isclass(call):
        return False
    call = getattr(call, "__call__", None)
    return inspect.iscoroutinefunction(call)


def is_async_gen_callable(call: Callable) -> TypeGuard[Callable[..., AsyncGenerator]]:
    if inspect.isasyncgenfunction(call):
        return True
    call = getattr(call, "__call__", None)
    return inspect.isasyncgenfunction(call)


def is_gen_callable(call: Callable) -> TypeGuard[Callable[..., Generator]]:
    if inspect.isgeneratorfunction(call):
        return True
    call = getattr(call, "__call__", None)
    return inspect.isgeneratorfunction(call)


def get_raw_handler(handler: Any) -> Any:
    """
    Get handler's raw handler.
    """
    while hasattr(handler, "__wrapped__") or hasattr(handler, "__raw_handler__"):
        new_handler = handler

        if hasattr(handler, "__wrapped__"):
            new_handler = handler.__wrapped__
        elif hasattr(handler, "__raw_handler__"):
            new_handler = handler.__raw_handler__

        if new_handler is handler:
            break

        handler = new_handler

    return handler


def get_object_filepath(object: Any) -> str:
    """
    Get object's filepath.
    """
    path = Path(inspect.getfile(object)).absolute()
    try:
        filepath = "./" + path.relative_to(Path.cwd()).as_posix()
    except ValueError:
        filepath = path.as_posix()
    return filepath



import inspect
import os
import sys
import typing

import click

from ..utils import import_from_string
from ..utils.inspect import get_object_filepath, get_raw_handler
from .extensions.multimethod import is_multimethod_view

if typing.TYPE_CHECKING:
    from indexpy import Index


@click.command(help="display all urls in application")
@click.argument("application")
def display_urls(application):
    sys.path.insert(0, os.getcwd())
    index_app: Index = import_from_string(application)
    for path, handler in index_app.router.http_tree.iterator():
        click.secho("* ", nl=False)
        click.secho(path, fg="green", nl=False)
        click.secho(" => ", nl=False)

        handler = get_raw_handler(handler)

        if is_multimethod_view(handler):
            click.secho("Is multi-method Endpoint")
            for method in handler.__methods__:
                func = get_raw_handler(getattr(handler, method.lower()))
                filepath = get_object_filepath(func)
                whitespaces = " " * (len(path) + len("* ") + len(" => "))
                click.secho(whitespaces + "| " + method + " => ", nl=False)
                click.secho(
                    filepath + ":" + str(func.__code__.co_firstlineno), fg="blue"
                )
        else:
            filepath = get_object_filepath(handler)
            click.secho(
                filepath + ":" + str(inspect.getsourcelines(handler)[1]), fg="blue"
            )



from __future__ import annotations

import functools
import inspect
from contextlib import _GeneratorContextManager, asynccontextmanager, contextmanager
from itertools import groupby
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from typing import cast as typing_cast

from baize.asgi import FormData
from pydantic import BaseConfig, BaseModel, ValidationError, create_model
from typing_extensions import Annotated, get_args, get_origin, get_type_hints

from ..concurrency import always_async
from ..exceptions import RequestValidationError
from ..requests import request
from ..utils import (
    is_async_gen_callable,
    is_coroutine_callable,
    is_gen_callable,
    safe_issubclass,
)
from .fields import DependInfo, FieldInfo, RequestInfo, Undefined

CallableObject = TypeVar("CallableObject", bound=Callable)


def create_model_config(title: str = None, description: str = None) -> Type[BaseConfig]:
    class ExclusiveModelConfig(BaseConfig):
        schema_extra = {
            k: v
            for k, v in {"title": title, "description": description}.items()
            if v is not None
        }

    return ExclusiveModelConfig


sorted_groupby = lambda iterable, key: groupby(sorted(iterable, key=key), key=key)


def _merge_multi_value(
    items: Sequence[Tuple[str, Any]]
) -> Dict[str, Union[Any, List[Any]]]:
    """
    If there are values with the same key value, they are merged into a List.
    """
    return {
        k: v_list if len(v_list) > 1 else v_list[0]
        for k, v_list in (
            (k, [v for _, v in kv_iter])
            for k, kv_iter in sorted_groupby(items, lambda kv: kv[0])
        )
    }


def create_new_callback(callback: CallableObject) -> CallableObject:
    sig = inspect.signature(callback)

    raw_parameters: Dict[str, Any] = {
        key: {} for key in ["path", "query", "header", "cookie", "body"]
    }
    exclusive_models: Dict[BaseModel, str] = {}

    for name, param in sig.parameters.items():
        if not (
            isinstance(param.default, FieldInfo)
            or (
                get_origin(param.annotation) is Annotated
                and isinstance(get_args(param.annotation)[1], FieldInfo)
            )
        ):
            continue

        if param.POSITIONAL_ONLY:
            raise TypeError(
                f"Parameter {name} cannot be defined as positional only parameters."
            )

        if isinstance(param.default, FieldInfo):
            type_ = param.annotation
            info = param.default
        else:
            type_, info = get_args(param.annotation)

        if getattr(info, "exclusive", False):
            if safe_issubclass(type_, BaseModel):
                model = type_
            else:
                model = create_model(
                    "temporary_exclusive_model",
                    __config__=create_model_config(info.title, info.description),
                    __root__=(type_, ...),
                )
            raw_parameters[info._in] = model
            exclusive_models[model] = name
        else:
            if safe_issubclass(raw_parameters[info._in], BaseModel):
                raise RuntimeError(
                    f"{info._in.capitalize()}(exclusive=True) "
                    f"and {info._in.capitalize()} cannot be used at the same time"
                )
            if type_ == param.empty:
                type_ = Any
            raw_parameters[info._in][name] = (type_, info)

    for key, params in filter(
        lambda kv: kv[1],
        ((key, raw_parameters.pop(key)) for key in tuple(raw_parameters.keys())),
    ):
        if safe_issubclass(params, BaseModel):
            model = params
        else:
            model = create_model("temporary_model", **params)  # type: ignore
        raw_parameters[key] = model

    if "body" in raw_parameters:
        request_body: Optional[BaseModel] = raw_parameters.pop("body")
    else:
        request_body = None

    if raw_parameters:
        parameters: Optional[Dict[str, BaseModel]] = typing_cast(
            Dict[str, BaseModel], raw_parameters
        )
    else:
        parameters = None

    request_attrs: Dict[str, RequestInfo] = {
        **{
            name: param.default
            for name, param in sig.parameters.items()
            if isinstance(param.default, RequestInfo)
        },
        **{
            name: get_args(param.annotation)[1]
            for name, param in sig.parameters.items()
            if get_origin(param.annotation) is Annotated
            and isinstance(get_args(param.annotation)[1], RequestInfo)
        },
    }

    depend_attrs: Dict[str, DependInfo] = {
        **{
            name: param.default
            for name, param in sig.parameters.items()
            if isinstance(param.default, DependInfo)
        },
        **{
            name: get_args(param.annotation)[1]
            for name, param in sig.parameters.items()
            if get_origin(param.annotation) is Annotated
            and isinstance(get_args(param.annotation)[1], DependInfo)
        },
    }

    if not (parameters or request_body or request_attrs or depend_attrs):
        return callback

    depend_functions = {
        name: create_new_callback(
            typing_cast(
                CallableObject, always_async(info.call) if info.to_async else info.call
            )
        )
        for name, info in depend_attrs.items()
    }

    @functools.wraps(callback)
    async def callback_with_auto_bound_params(*args, **kwargs):
        data: List[Any] = []
        keyword_params: Dict[str, Any] = {}

        # try to call depend functions
        need_closes = []
        for name, function in depend_functions.items():
            info = depend_attrs[name]
            if is_async_gen_callable(info.call):
                asyncgenerator = asynccontextmanager(function)()
                if inspect.isawaitable(asyncgenerator.gen):
                    asyncgenerator.gen = await asyncgenerator.gen
                keyword_params[name] = await asyncgenerator.__aenter__()
                need_closes.append(asyncgenerator)
            elif is_coroutine_callable(info.call):
                keyword_params[name] = await function()
            elif is_gen_callable(info.call):
                if info.to_async:
                    asyncgenerator = asynccontextmanager(function)()
                    if inspect.isawaitable(asyncgenerator.gen):
                        asyncgenerator.gen = await asyncgenerator.gen
                    keyword_params[name] = await asyncgenerator.__aenter__()
                    need_closes.append(asyncgenerator)
                else:
                    generator = contextmanager(function)()
                    if inspect.isawaitable(generator.gen):
                        generator.gen = await generator.gen
                    keyword_params[name] = generator.__enter__()
                    need_closes.append(generator)
            else:
                result = function()
                if inspect.isawaitable(result):
                    result = await result
                keyword_params[name] = result

        # try to get parameters model and parse
        if parameters:
            if "path" in parameters:
                try:
                    data.append(parameters["path"].parse_obj(request.path_params))
                except ValidationError as e:
                    raise RequestValidationError(e, "path")

            if "query" in parameters:
                try:
                    data.append(
                        parameters["query"].parse_obj(
                            _merge_multi_value(request.query_params.multi_items())
                        )
                    )
                except ValidationError as e:
                    raise RequestValidationError(e, "query")

            if "header" in parameters:
                try:
                    data.append(parameters["header"].parse_obj(request.headers))
                except ValidationError as e:
                    raise RequestValidationError(e, "header")

            if "cookie" in parameters:
                try:
                    data.append(parameters["cookie"].parse_obj(request.cookies))
                except ValidationError as e:
                    raise RequestValidationError(e, "cookie")

        # try to get body model and parse
        if request_body:
            _body_data = await request.data()
            if isinstance(_body_data, FormData):
                _body_data = _merge_multi_value(_body_data.multi_items())
            try:
                data.append(request_body.parse_obj(_body_data))
            except ValidationError as e:
                raise RequestValidationError(e, "body")

        # try to get request instance attributes
        if request_attrs:
            for name, info in request_attrs.items():
                try:
                    value: Any = functools.reduce(
                        lambda attr, name: getattr(attr, name),
                        (info.alias or name).split("."),
                        request,
                    )
                except AttributeError:
                    if info.default is not Undefined:
                        value = info.default
                    elif info.default_factory is not None:
                        value = info.default_factory()
                    else:
                        raise
                keyword_params[name] = value

        for _data in data:
            if _data.__class__.__name__ == "temporary_model":
                keyword_params.update(_data.dict())
            elif _data.__class__.__name__ == "temporary_exclusive_model":
                keyword_params[exclusive_models[_data.__class__]] = _data.__root__
            else:
                keyword_params[exclusive_models[_data.__class__]] = _data

        try:
            result = callback(*args, **{**keyword_params, **kwargs})  # type: ignore
            if inspect.isawaitable(result):
                result = await result
            return result
        finally:
            for need_close in need_closes:
                if isinstance(need_close, _GeneratorContextManager):
                    need_close.__exit__(None, None, None)
                else:
                    await need_close.__aexit__(None, None, None)

    handler = callback_with_auto_bound_params

    if request_body is not None:
        __request_body__: List[BaseModel] = getattr(
            handler, "__docs_request_body__", []
        )
        __request_body__.append(request_body)
        setattr(handler, "__docs_request_body__", __request_body__)

    if parameters is not None:
        __parameters__: Dict[str, List[BaseModel]] = getattr(
            handler, "__docs_parameters__", {}
        )
        for key, parameter_model in parameters.items():
            __parameters__.setdefault(key, []).append(parameter_model)
        setattr(handler, "__docs_parameters__", __parameters__)

    for func in depend_functions.values():
        __request_body__ = getattr(handler, "__docs_request_body__", [])
        __request_body__.extend(getattr(func, "__docs_request_body__", []))
        setattr(handler, "__docs_request_body__", __request_body__)

        __parameters__ = getattr(handler, "__docs_parameters__", {})
        for key, value in getattr(func, "__docs_parameters__", {}).items():
            __parameters__.setdefault(key, []).extend(value)
        setattr(handler, "__docs_parameters__", __parameters__)

    __signature__ = inspect.Signature(
        parameters=[
            param
            for param in sig.parameters.values()
            if not (
                isinstance(param.default, (FieldInfo, RequestInfo, DependInfo))
                or (
                    get_origin(param.annotation) is Annotated
                    and isinstance(
                        get_args(param.annotation)[1], (FieldInfo, RequestInfo)
                    )
                )
            )
        ],
        return_annotation=sig.return_annotation,
    )
    setattr(callback_with_auto_bound_params, "__signature__", __signature__)

    del callback_with_auto_bound_params.__wrapped__  # type: ignore

    return callback_with_auto_bound_params  # type: ignore


def create_new_class(cls):
    """
    Create a fake class for auto-bound parameters.
    """

    @functools.wraps(cls, updated=())
    class NewClass(cls):
        pass

    return NewClass


def auto_params(handler: CallableObject) -> CallableObject:
    if inspect.isclass(handler) and hasattr(handler, "__methods__"):
        new_class = create_new_class(handler)
        for method in map(lambda x: x.lower(), handler.__methods__):  # type: ignore
            old_callback = getattr(handler, method)
            new_callback = create_new_callback(old_callback)
            setattr(new_class, method, new_callback)  # note: set to new class
            setattr(
                new_callback, "__docs_responses__", parse_docs_responses(old_callback)
            )
        setattr(new_class, "__raw_handler__", handler)
        return new_class
    elif inspect.iscoroutinefunction(handler) or (
        callable(handler) and inspect.iscoroutinefunction(handler.__call__)  # type: ignore
    ):
        old_callback = handler
        new_callback = create_new_callback(handler)
        setattr(new_callback, "__docs_responses__", parse_docs_responses(old_callback))
        setattr(new_callback, "__raw_handler__", handler)
        return new_callback
    else:
        return handler


def update_wrapper(
    new_handler: CallableObject, old_handler: Callable[..., Awaitable[Any]]
) -> CallableObject:
    """
    Update wrapper for auto-bound parameters.
    """
    for attr in dir(old_handler):
        if attr.startswith("__docs_") or attr in ("__method__", "__methods__"):
            setattr(new_handler, attr, getattr(old_handler, attr))

    return new_handler


def parse_docs_responses(callback):
    return {
        **getattr(callback, "__docs_responses__", {}),
        **functools.reduce(
            lambda x, y: {**x, **y},
            (
                response
                for response in get_args(
                    (
                        get_type_hints(callback.__call__, include_extras=True)
                        if (callable(callback) and not inspect.isfunction(callback))
                        else get_type_hints(callback, include_extras=True)
                    ).get("return")
                )
                if isinstance(response, dict)
            ),
            {},
        ),
    }
